@@ -171,11 +171,11 @@ gtk_grid_get_property (GObject    *object,
       break;
 
     case PROP_ROW_HOMOGENEOUS:
-      g_value_set_boolean (value, ROWS (priv)->homogeneous);
+      g_value_set_boolean (value, COLUMNS (priv)->homogeneous);
       break;
 
     case PROP_COLUMN_HOMOGENEOUS:
-      g_value_set_boolean (value, COLUMNS (priv)->homogeneous);
+      g_value_set_boolean (value, ROWS (priv)->homogeneous);
       break;
 
     default:
@@ -458,17 +458,6 @@ gtk_grid_child_type (GtkContainer *container)
   return GTK_TYPE_WIDGET;
 }
 
-static GtkSizeRequestMode
-gtk_grid_get_request_mode (GtkWidget *widget)
-{
-  GtkGridPrivate *priv = GTK_GRID (widget)->priv;
-
-  if (priv->orientation == GTK_ORIENTATION_VERTICAL)
-    return GTK_SIZE_REQUEST_HEIGHT_FOR_WIDTH;
-  else
-    return GTK_SIZE_REQUEST_WIDTH_FOR_HEIGHT;
-}
-
 /* Calculates the min and max numbers for both orientations.
  */
 static void
@@ -1059,6 +1048,15 @@ gtk_grid_get_size (GtkGrid        *grid,
   GtkGridRequest request;
   GtkGridLines *lines;
 
+  if (minimum)
+    *minimum = 0;
+
+  if (natural)
+    *natural = 0;
+
+  if (grid->priv->children == NULL)
+    return;
+
   request.grid = grid;
   gtk_grid_request_count_lines (&request);
   lines = &request.lines[orientation];
@@ -1080,6 +1078,15 @@ gtk_grid_get_size_for_size (GtkGrid        *grid,
   GtkGridLines *lines;
   gint min_size;
 
+  if (minimum)
+    *minimum = 0;
+
+  if (natural)
+    *natural = 0;
+
+  if (grid->priv->children == NULL)
+    return;
+
   request.grid = grid;
   gtk_grid_request_count_lines (&request);
   lines = &request.lines[0];
@@ -1104,8 +1111,7 @@ gtk_grid_get_preferred_width (GtkWidget *widget,
 {
   GtkGrid *grid = GTK_GRID (widget);
 
-  if (gtk_grid_get_request_mode (widget) == GTK_SIZE_REQUEST_WIDTH_FOR_HEIGHT)
-
+  if (gtk_widget_get_request_mode (widget) == GTK_SIZE_REQUEST_WIDTH_FOR_HEIGHT)
     gtk_grid_get_size_for_size (grid, GTK_ORIENTATION_HORIZONTAL, 0, minimum, natural);
   else
     gtk_grid_get_size (grid, GTK_ORIENTATION_HORIZONTAL, minimum, natural);
@@ -1118,7 +1124,7 @@ gtk_grid_get_preferred_height (GtkWidget *widget,
 {
   GtkGrid *grid = GTK_GRID (widget);
 
-  if (gtk_grid_get_request_mode (widget) == GTK_SIZE_REQUEST_HEIGHT_FOR_WIDTH)
+  if (gtk_widget_get_request_mode (widget) == GTK_SIZE_REQUEST_HEIGHT_FOR_WIDTH)
     gtk_grid_get_size_for_size (grid, GTK_ORIENTATION_VERTICAL, 0, minimum, natural);
   else
     gtk_grid_get_size (grid, GTK_ORIENTATION_VERTICAL, minimum, natural);
@@ -1132,7 +1138,7 @@ gtk_grid_get_preferred_width_for_height (GtkWidget *widget,
 {
   GtkGrid *grid = GTK_GRID (widget);
 
-  if (gtk_grid_get_request_mode (widget) == GTK_SIZE_REQUEST_WIDTH_FOR_HEIGHT)
+  if (gtk_widget_get_request_mode (widget) == GTK_SIZE_REQUEST_WIDTH_FOR_HEIGHT)
     gtk_grid_get_size_for_size (grid, GTK_ORIENTATION_HORIZONTAL, height, minimum, natural);
   else
     gtk_grid_get_size (grid, GTK_ORIENTATION_HORIZONTAL, minimum, natural);
@@ -1146,7 +1152,7 @@ gtk_grid_get_preferred_height_for_width (GtkWidget *widget,
 {
   GtkGrid *grid = GTK_GRID (widget);
 
-  if (gtk_grid_get_request_mode (widget) == GTK_SIZE_REQUEST_HEIGHT_FOR_WIDTH)
+  if (gtk_widget_get_request_mode (widget) == GTK_SIZE_REQUEST_HEIGHT_FOR_WIDTH)
     gtk_grid_get_size_for_size (grid, GTK_ORIENTATION_VERTICAL, width, minimum, natural);
   else
     gtk_grid_get_size (grid, GTK_ORIENTATION_VERTICAL, minimum, natural);
@@ -1207,6 +1213,10 @@ gtk_grid_request_allocate_children (GtkGridRequest *request)
       child_allocation.width = MAX (1, width);
       child_allocation.height = MAX (1, height);
 
+      if (gtk_widget_get_direction (GTK_WIDGET (request->grid)) == GTK_TEXT_DIR_RTL)
+        child_allocation.x = allocation.x + allocation.width
+                             - (child_allocation.x - allocation.x) - child_allocation.width;
+
       gtk_widget_size_allocate (child->widget, &child_allocation);
     }
 }
@@ -1221,6 +1231,13 @@ gtk_grid_size_allocate (GtkWidget     *widget,
   GtkGridPrivate *priv = grid->priv;
   GtkGridRequest request;
   GtkGridLines *lines;
+  GtkOrientation orientation;
+
+  if (priv->children == NULL)
+    {
+      gtk_widget_set_allocation (widget, allocation);
+      return;
+    }
 
   request.grid = grid;
 
@@ -1234,10 +1251,15 @@ gtk_grid_size_allocate (GtkWidget     *widget,
 
   gtk_widget_set_allocation (widget, allocation);
 
-  gtk_grid_request_run (&request, 1 - priv->orientation, FALSE);
-  gtk_grid_request_allocate (&request, 1 - priv->orientation, GET_SIZE (allocation, 1 - priv->orientation));
-  gtk_grid_request_run (&request, priv->orientation, TRUE);
-  gtk_grid_request_allocate (&request, priv->orientation, GET_SIZE (allocation, priv->orientation));
+  if (gtk_widget_get_request_mode (widget) == GTK_SIZE_REQUEST_WIDTH_FOR_HEIGHT)
+    orientation = GTK_ORIENTATION_HORIZONTAL;
+  else
+    orientation = GTK_ORIENTATION_VERTICAL;
+
+  gtk_grid_request_run (&request, 1 - orientation, FALSE);
+  gtk_grid_request_allocate (&request, 1 - orientation, GET_SIZE (allocation, 1 - orientation));
+  gtk_grid_request_run (&request, orientation, TRUE);
+  gtk_grid_request_allocate (&request, orientation, GET_SIZE (allocation, orientation));
 
   gtk_grid_request_position (&request, 0);
   gtk_grid_request_position (&request, 1);
@@ -1258,7 +1280,6 @@ gtk_grid_class_init (GtkGridClass *class)
   widget_class->size_allocate = gtk_grid_size_allocate;
   widget_class->get_preferred_width = gtk_grid_get_preferred_width;
   widget_class->get_preferred_height = gtk_grid_get_preferred_height;
-  widget_class->get_request_mode = gtk_grid_get_request_mode;
   widget_class->get_preferred_width_for_height = gtk_grid_get_preferred_width_for_height;
   widget_class->get_preferred_height_for_width = gtk_grid_get_preferred_height_for_width;
 
@@ -1458,6 +1479,178 @@ gtk_grid_attach_next_to (GtkGrid         *grid,
   grid_attach (grid, child, left, top, width, height);
 }
 
+/**
+ * gtk_grid_get_child_at:
+ * @grid: a #GtkGrid
+ * @left: the left edge of the cell
+ * @top: the top edge of the cell
+ *
+ * Gets the child of @grid whose area covers the grid
+ * cell whose upper left corner is at @left, @top.
+ *
+ * Returns: the child at the given position, or %NULL
+ *
+ * Since: 3.2
+ */
+GtkWidget *
+gtk_grid_get_child_at (GtkGrid *grid,
+                       gint     left,
+                       gint     top)
+{
+  GtkGridPrivate *priv = grid->priv;
+  GtkGridChild *child;
+  GList *list;
+
+  for (list = priv->children; list; list = list->next)
+    {
+      child = list->data;
+
+      if (CHILD_LEFT (child) <= left &&
+          CHILD_LEFT (child) + CHILD_WIDTH (child) > left &&
+          CHILD_TOP (child) <= top &&
+          CHILD_TOP (child) + CHILD_HEIGHT (child) > top)
+        return child->widget;
+    }
+
+  return NULL;
+}
+
+/**
+ * gtk_grid_insert_row:
+ * @grid: a #GtkGrid
+ * @position: the position to insert the row at
+ *
+ * Inserts a row at the specified position.
+ *
+ * Children which are attached at or below this position
+ * are moved one row down. Children which span across this
+ * position are grown to span the new row.
+ *
+ * Since: 3.2
+ */
+void
+gtk_grid_insert_row (GtkGrid *grid,
+                     gint     position)
+{
+  GtkGridPrivate *priv = grid->priv;
+  GtkGridChild *child;
+  GList *list;
+  gint top, height;
+
+  g_return_if_fail (GTK_IS_GRID (grid));
+
+  for (list = priv->children; list; list = list->next)
+    {
+      child = list->data;
+
+      top = CHILD_TOP (child);
+      height = CHILD_HEIGHT (child);
+
+      if (top >= position)
+        {
+          CHILD_TOP (child) = top + 1;
+          gtk_container_child_notify (GTK_CONTAINER (grid), child->widget, "top-attach");
+        }
+      else if (top + height > position)
+        {
+          CHILD_HEIGHT (child) = height + 1;
+          gtk_container_child_notify (GTK_CONTAINER (grid), child->widget, "height");
+        }
+    }
+}
+
+/**
+ * gtk_grid_insert_column:
+ * @grid: a #GtkGrid
+ * @position: the position to insert the column at
+ *
+ * Inserts a column at the specified position.
+ *
+ * Children which are attached at or to the right of this position
+ * are moved one column to the right. Children which span across this
+ * position are grown to span the new column.
+ *
+ * Since: 3.2
+ */
+void
+gtk_grid_insert_column (GtkGrid *grid,
+                        gint     position)
+{
+  GtkGridPrivate *priv = grid->priv;
+  GtkGridChild *child;
+  GList *list;
+  gint left, width;
+
+  g_return_if_fail (GTK_IS_GRID (grid));
+
+  for (list = priv->children; list; list = list->next)
+    {
+      child = list->data;
+
+      left = CHILD_LEFT (child);
+      width = CHILD_WIDTH (child);
+
+      if (left >= position)
+        {
+          CHILD_LEFT (child) = left + 1;
+          gtk_container_child_notify (GTK_CONTAINER (grid), child->widget, "left-attach");
+        }
+      else if (left + width > position)
+        {
+          CHILD_WIDTH (child) = width + 1;
+          gtk_container_child_notify (GTK_CONTAINER (grid), child->widget, "width");
+        }
+    }
+}
+
+/**
+ * gtk_grid_insert_next_to:
+ * @grid: a #GtkGrid
+ * @sibling: the child of @grid that the new row or column will be
+ *     placed next to
+ * @side: the side of @sibling that @child is positioned next to
+ *
+ * Inserts a row or column at the specified position.
+ *
+ * The new row or column is placed next to @sibling, on the side
+ * determined by @side. If @side is %GTK_POS_TOP or %GTK_POS_BOTTOM,
+ * a row is inserted. If @side is %GTK_POS_LEFT of %GTK_POS_RIGHT,
+ * a column is inserted.
+ *
+ * Since: 3.2
+ */
+void
+gtk_grid_insert_next_to (GtkGrid         *grid,
+                         GtkWidget       *sibling,
+                         GtkPositionType  side)
+{
+  GtkGridChild *child;
+
+  g_return_if_fail (GTK_IS_GRID (grid));
+  g_return_if_fail (GTK_IS_WIDGET (sibling));
+  g_return_if_fail (gtk_widget_get_parent (sibling) == (GtkWidget*)grid);
+
+  child = find_grid_child (grid, sibling);
+
+  switch (side)
+    {
+    case GTK_POS_LEFT:
+      gtk_grid_insert_column (grid, CHILD_LEFT (child));
+      break;
+    case GTK_POS_RIGHT:
+      gtk_grid_insert_column (grid, CHILD_LEFT (child) + CHILD_WIDTH (child));
+      break;
+    case GTK_POS_TOP:
+      gtk_grid_insert_row (grid, CHILD_TOP (child));
+      break;
+    case GTK_POS_BOTTOM:
+      gtk_grid_insert_row (grid, CHILD_TOP (child) + CHILD_HEIGHT (child));
+      break;
+    default:
+      g_assert_not_reached ();
+    }
+}
+
 /**
  * gtk_grid_set_row_homogeneous:
  * @grid: a #GtkGrid
@@ -1474,9 +1667,10 @@ gtk_grid_set_row_homogeneous (GtkGrid  *grid,
 
   priv = grid->priv;
 
-  if (ROWS (priv)->homogeneous != homogeneous)
+  /* Yes, homogeneous rows means all the columns have the same size */
+  if (COLUMNS (priv)->homogeneous != homogeneous)
     {
-      ROWS (priv)->homogeneous = homogeneous;
+      COLUMNS (priv)->homogeneous = homogeneous;
 
       if (gtk_widget_get_visible (GTK_WIDGET (grid)))
         gtk_widget_queue_resize (GTK_WIDGET (grid));
@@ -1501,7 +1695,7 @@ gtk_grid_get_row_homogeneous (GtkGrid *grid)
 
   priv = grid->priv;
 
-  return ROWS (priv)->homogeneous;
+  return COLUMNS (priv)->homogeneous;
 }
 
 /**
@@ -1520,9 +1714,10 @@ gtk_grid_set_column_homogeneous (GtkGrid  *grid,
 
   priv = grid->priv;
 
-  if (COLUMNS (priv)->homogeneous != homogeneous)
+  /* Yes, homogeneous columns means all the rows have the same size */
+  if (ROWS (priv)->homogeneous != homogeneous)
     {
-      COLUMNS (priv)->homogeneous = homogeneous;
+      ROWS (priv)->homogeneous = homogeneous;
 
       if (gtk_widget_get_visible (GTK_WIDGET (grid)))
         gtk_widget_queue_resize (GTK_WIDGET (grid));
@@ -1547,7 +1742,7 @@ gtk_grid_get_column_homogeneous (GtkGrid *grid)
 
   priv = grid->priv;
 
-  return COLUMNS (priv)->homogeneous;
+  return ROWS (priv)->homogeneous;
 }
 
 /**
